
/* GLUT/GL/Texture Demonstrator Program
 * Leonard D. Brown
 *
 * Acknowledgements:
 *		Portions of demo code provided by Kobus Barnard,
 *		University of Arizona
 * 
 *		ByteRaster class and methods modified from Michael Garland's
 *		LibGFX library in accordance with granted terms and conditions
 *		(see http://graphics.cs.uiuc.edu/~garland/software/libgfx.html)
 *
 * To make this code relatively independent of build paths and install
 *      paths, I've based directory paths in the VS project on environment 
 *		variables GLUT_PATH and LIBJPEG_PATH.  These environment variables
 *		must be defined by the OS should point to the root directories of 
 *		the GLUT and LIBJPEG libraries, respectively.
 *
 * If your system is lacking these APIs, you can find them here:
 *		GLUT: http://www.xmission.com/~nate/glut.html
 *	    LibJPEG: Linux: http://www.ijg.org/, 
 *				 Windows: Use the library provided on the CSc433 website OpenGL page;
 *						  other versions may cause jpeg_read_header to crash due 
 *					      to standard C library incompatibility issue with MSVC
 *						  unless you rebuild the entire library from source
 */

#include <setjmp.h>
#include <stdio.h>       
#include <stdlib.h>       

#if defined(WIN32)
#  include <windows.h>
#endif

/* RGB channel definitions */
#define RED   0
#define GREEN 1
#define BLUE  2

#define WHITE 255

#define INCREMENT 1.0

/*
// Sometimes glu.h includes glut.h--sometimes not.
*/
#ifdef MAC_OSX
#	 include <OpenGL/glew.h>
#    include <OpenGL/glu.h>
#    include <GLUT/glut.h>
#	 include <sys/stat.h>
#else 
	#ifdef WIN32
	#	 include <GL/glew.h>
	#	 include <GL/glu.h>
	#	 include <glut.h>
	#else 
	#	 include <GL/glew.h>
	#    include <GL/glu.h>    
	#    include <GL/glut.h>       
	#endif
#endif

#include <memory.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "raster.h"

/*
// Some debugging macros. Normally I make macros upper case; these are
// exceptions. 
*/

#define dbc(X)                                                                 \
    if (debug_level >= 1)                                                      \
    {                                                                          \
        fprintf(stderr,                                                        \
                "dbc: "#X" is ->%c<- (0x%x, %d) on source line %d of %s\n",    \
                (int)(X), (int)(X), (int)(X), __LINE__, __FILE__);             \
    }   

#define dbp(X) \
    if (debug_level >= 1)                                                      \
    {                                                                          \
        fprintf(stderr, "%s\n", (X));                                          \
    }   

#define dbi(X)                                                                 \
    if (debug_level >= 1)                                                      \
    {                                                                          \
        fprintf(stderr,                                                        \
                "dbi: "#X" is %ld on source line %d of %s\n", (long)(X),       \
               __LINE__, __FILE__);                                            \
    }   

#define  dbw()                                                                 \
    if (debug_level >= 1)                                                      \
    {                                                                          \
        fprintf(stderr,                                                        \
                "At line %d of %s\n", __LINE__, __FILE__);                     \
    }   


/* -------------------------------------------------------------------------- */

#define NOT_SET  (-99)     /* Something negative. */

/* -------------------------------------------------------------------------- */

/* 
// Some prefer to write their programs backwards, so we don't need to declare
// functions up front--I prefer to write top down, so I have added function
// declarations to the original program. 
*/

static void display_CB(void);            /* Called whenever redisplay needed */
static void idle_CB   (void);
static void key_CB(unsigned char key, int x, int y);     /* Called on key press */

/* -------------------------------------------------------------------------- */

static int debug_level = 1;  /* Used by the macros above. */


/* -------------------------------------------------------------------------- */

/* window origin on our display device, in display (pixel) coordinates */
static int x_orig = 0;
static int y_orig = 0;
/* default window size on our display device, in pixels */
static int width  = 640;
static int height = 480;

static float angle = 0.0;

/* animation flags */
static int anime_R = 0;
static int anime_G = 0; 
static int anime_B = 0;

/* mouse */
static int drag_start_x;
static int drag_start_y;
static int displacement_x;
static int displacement_y; 

/* brush color intensity */
static float color[3] = {0.4, 0.4, 0.4};

/* some hard-coded image file paths */
static char infile[]  = "pacific.jpg";
static char outfile[] = "output.jpg";

/* a raster scan of the image   */
static ByteRaster *image;
/* handles to shader programs   */
GLuint shaderProg[2];

/* -------------------------------------------------------------------------- */
/* PRINT THE DEBUG LOG:
//   Prints a status log for a vertex shader, fragment shader, 
//   or shader program.  Indicates compile and linking errors
//	 encountered at run time.  Useful for debugging. This code
//   adapted from samples at gpwiki.org (now wiki.gamedev.net).
*/

void printLog (GLuint obj)
{
	int   infologLength = 0;
	GLint maxLength;
	
	if (glIsShader (obj))
		glGetShaderiv (obj, GL_INFO_LOG_LENGTH, &maxLength);
	else
		glGetProgramiv(obj, GL_INFO_LOG_LENGTH, &maxLength);
			
	char *infoLog = new char[maxLength];
 
	if (glIsShader(obj))
		glGetShaderInfoLog (obj, maxLength, &infologLength, infoLog);
	else
		glGetProgramInfoLog(obj, maxLength, &infologLength, infoLog);
 
	if (infologLength > 0)
		printf("%s\n",infoLog);
}

/* -------------------------------------------------------------------------- */
/* LOADING A SHADER
//   local helper function for loading a shader from a 
//   file.  This source code was adapted from Yong Cao's
//	 example slides.
*/

char* readShaderSource (const char* shaderFile)
{
	struct stat statBuf;
	FILE* fp = fopen(shaderFile, "r");
	char* buf;

	fstat (fileno(fp), &statBuf);
	buf = (char*) malloc ((statBuf.st_size + 1) * sizeof(char));
	memset (buf, '\0', statBuf.st_size + 1);

	fread (buf, 1, (size_t) statBuf.st_size, fp);
	fclose (fp);

	return buf;
}

/* -------------------------------------------------------------------------- */
/* BUILDING A SHADER: 
//   construct a shader program from constituent vertex
//	 and fragment programs loaded from the file paths
//	 vShaderFile and fShaderFile, respectively.
*/

void build_shader (GLuint shaderHandle, char *vertShaderFile, char *fragShaderFile)
{
	GLuint vertHandle;
	GLuint fragHandle;

	const GLchar *vertSource, 
				 *fragSource;

	/* load, compile, and attach the vertex shader */
	vertSource = readShaderSource (vertShaderFile);
	vertHandle = glCreateShader(GL_VERTEX_SHADER);
	glShaderSource (vertHandle, 1, &vertSource, NULL);
	glCompileShader(vertHandle);
	glAttachShader (shaderHandle, vertHandle);

	fprintf (stdout, "Vertex Shader code:\n%s\n", vertSource );
	printLog(vertHandle);

	/* load, compile, and attach the fragment shader */
	fragSource = readShaderSource (fragShaderFile);
	fragHandle = glCreateShader(GL_FRAGMENT_SHADER);
	glShaderSource (fragHandle, 1, &fragSource, NULL);
	glCompileShader(fragHandle);
	glAttachShader (shaderHandle, fragHandle);

	fprintf (stdout, "Fragment Shader code:\n%s\n", fragSource );
	printLog(fragHandle);
}

/* -------------------------------------------------------------------------- */
/* INITIALIZE THE TEXTURE:													  
//	 This function performs a number of steps necessary						
//   to initialize the GL texture map and load it to the 
//   hardware.  You can largely ignore this function and
//   copy it verbatim if you want to use hardware texture
//   support in your program.  Note: this function is not
//   robust, so isn't recommended for loading lots of 
//   textures repeatedly.
*/

void init_texture (ByteRaster &img)
{
	img.vflip();

	GLuint handle;

	glGenTextures (1, &handle);

	glBindTexture (GL_TEXTURE_2D, handle);

	glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,	   
					 GL_REPEAT);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T,	   
					 GL_REPEAT);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, 
					 GL_LINEAR); 

	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, 
					 GL_LINEAR_MIPMAP_LINEAR);

	gluBuild2DMipmaps(GL_TEXTURE_2D,
					  GL_RGB,
					  img.width(), img.height(), 
					  GL_RGB,
					  GL_UNSIGNED_BYTE,
					  img.head());
}

/* draw a wireframe unit cube with origin at centroid */
void draw_cube (void)
{
	glColor3d (0,1,0);

	glBegin (GL_LINES);
		/* front bottom */
		glVertex3d (-0.5,-0.5, 0.5);
		glVertex3d ( 0.5,-0.5, 0.5);
		/* front right */
		glVertex3d ( 0.5,-0.5, 0.5);
		glVertex3d ( 0.5, 0.5, 0.5);
		/* front top */
		glVertex3d ( 0.5, 0.5, 0.5);
		glVertex3d (-0.5, 0.5, 0.5);
		/* front left */
		glVertex3d (-0.5, 0.5, 0.5);
		glVertex3d (-0.5,-0.5, 0.5);
		/* top left */
		glVertex3d (-0.5, 0.5, 0.5);
		glVertex3d (-0.5, 0.5,-0.5);
		/* top right */
		glVertex3d ( 0.5, 0.5, 0.5);
		glVertex3d ( 0.5, 0.5,-0.5);
		/* bottom left */
		glVertex3d (-0.5,-0.5, 0.5);
		glVertex3d (-0.5,-0.5,-0.5);
		/* bottom right */
		glVertex3d ( 0.5,-0.5, 0.5);
		glVertex3d ( 0.5,-0.5,-0.5);
		/* back bottom */
		glVertex3d (-0.5,-0.5,-0.5);
		glVertex3d ( 0.5,-0.5,-0.5);
		/* back right */
		glVertex3d ( 0.5,-0.5,-0.5);
		glVertex3d ( 0.5, 0.5,-0.5);
		/* back top */
		glVertex3d ( 0.5, 0.5,-0.5);
		glVertex3d (-0.5, 0.5,-0.5);
		/* back left */
		glVertex3d (-0.5, 0.5,-0.5);
		glVertex3d (-0.5,-0.5,-0.5);
	glEnd ();
}

/* -------------------------------------------------------------------------- */
int main(int argc, char *argv[])
{
	/* window handle */
    int win;
	GLenum glewErr;

	/* Read an RGB byte raster (of dimension w x h x 3) for an image */
	image = read_jpeg_image(infile);

	/* set the window dimensions to exactly fit the image. */
	if (image)
	{
		//width  = image->width ();
		//height = image->height();
	}

	/*-------------------------------------------------------------*/
	/* GLUT & GL initialization */

	/* initialize GLUT system */
    glutInit(&argc, argv);    
	/* initialize display format */
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);

	/* define and get handle to the window (render) context */
    glutInitWindowSize(width, height); 
    win = glutCreateWindow("Demo"); 

	/* init extension wrangler library; this *must* occur after we	*/
	/* acquired a valid Open GL rendering context from glutInit		*/
	glewErr = glewInit ();
	if (glewErr != GLEW_OK)
	{
	  /* Problem: glewInit failed, something is seriously wrong. */
	  fprintf(stderr, "Error: %s\n", glewGetErrorString (glewErr));
	}
	fprintf (stdout, "Status: Using GLEW %s\n", glewGetString(GLEW_VERSION));

	if (!GLEW_VERSION_2_0) 	
	{
		fprintf (stdout, "Error: Required extensions not supported.\n");
        fprintf (stdout, "  Program requires OpenGL version 2.0 or later.\n");
        exit   (-1);
    }

	/* set window's display callback */
    glutDisplayFunc(display_CB);       
	/* background events and animation */
	glutIdleFunc (idle_CB);
	/* set window's key callback */
    glutKeyboardFunc(key_CB);

	/* set window's mouse callback */
	/*
    //glutMouseFunc(mouse_CB); 
	*/
	/* set window's mouse move with button pressed callback */
	/*
    //glutMotionFunc(mouse_move_CB);      
	*/

    /* set background clear color to black */
    glClearColor((GLclampf)0.0,(GLclampf)0.0,(GLclampf)0.0,(GLclampf)0.0);  

	glEnable     (GL_DEPTH_TEST);


	/* end GLUT, GLEW, & GL initialization */
	/*-------------------------------------------------------------*/

	/* setup the perspective projection */
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();

	/* mapping of the points on the viewplane (NDC) to window     */
	glViewport (0,0,640,480);
	/* specify the view volume, alternatively may use glFrustum() */
	/* for PERSPECTIVE projection. For ORTHOGRAPHIC projection,   */
	/* use glOrtho() given below								  */
	gluPerspective (45.0,1.333,1.0,1000.0);
	/* glOrtho (-320, 320, -240, 240, 1, 1000.0); */
	
	/* construct the shader programs. */
	shaderProg[0] = glCreateProgram ();
	shaderProg[1] = glCreateProgram ();

	build_shader (shaderProg[0], "vertex1.glsl", "fragment1.glsl");
	build_shader (shaderProg[1], "vertex2.glsl", "fragment2.glsl");

	glLinkProgram (shaderProg[0]);
	glLinkProgram (shaderProg[1]);

	printLog(shaderProg[0]);
	printLog(shaderProg[1]);

	init_texture (*image);

	/* start processing events...this is the event 'pump' */
    glutMainLoop();                      

    /* Execution never reaches this point, so return value is failure */
    return EXIT_FAILURE;
}


/* -------------------------------------------------------------------------- */
/*
// Function called whenever redisplay needed (try hiding the window and then
// exposing it. The debug macro should print something.)
*/
static void display_CB(void)
{
	/* this draw callback demonstrates some common GL functionality:		*/
	/* draws 2 polygons, a polygon frame, and some pixels into the buffer	*/
	/* at arbitrary locations												*/

	/* clear the drawing buffer */
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);       

	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();

	/* configure the shader texture support:						   */
	/* get a handle to the texture image handle declared in the shader */
	GLint mapHandle = glGetUniformLocation (shaderProg[1], "image");
	/* map the texture image handle to the hardware texture unit where */
	/* the texture is currently loaded (tex unit 0 by default)		   */
	glUniform1i (mapHandle, 0);

	/* specify the world-to-camera transform */
	gluLookAt (0,0,600, 0,0,0, 0,1,0);

	/* increment rotation component */
	angle += INCREMENT;

	/* use the custom shader program pointed to by the given handle,	  */
	/* until we call this function again to change it					  */	
	glUseProgram (shaderProg[1]);
	glPushMatrix ();
	{
		/* the shader program's vertex shader has a global attribute	  */
		/* that is used to maintain texture coordinates.  get a handle    */
		/* to it.  Note: there are built-in vertex attributes that are	  */
		/* used for managing texture coordinates in standard practice.	  */
		GLint uvcoord = glGetAttribLocation (shaderProg[1], "vertex_texture_coordinate");
		GLfloat uv[2];

		/* this polygon is 100 units in front of the image */
		glTranslatef (0,100,100);

		/* rotate about centroid of polygon, on y axis */
		glRotatef (-angle, 0,1,0);

		/* note on texture coordinates:									  */
		/*	 Texture coordinates are assigned such that (0,0) designates  */
		/*   the left bottom corner of the image and (1,1) designates the */
		/*   upper right corner. Floats >1.0 will cause the image to wrap */
		/*   around in the current configuration.  Visually, this will	  */
		/*   cause the texture to repeat.								  */

		glBegin (GL_QUADS);
			uv[0] = 0; uv[1] = 0;
			/* communicate the uv coordinates associated with each vertex */
			/* to the shader program's variable; the uv coordinates for   */
			/* pixels between the vertices are interpolated.			  */
			glVertexAttrib2fv(uvcoord, uv);
			glColor3f  (1.0,0.5,0.2);
			glVertex3d (-50,-50,0);

			uv[0] = 2; uv[1] = 0;
			glVertexAttrib2fv(uvcoord, uv);
			glColor3f  (1.0,0.5,0.2);
			glVertex3d ( 50,-50,0);

			uv[0] = 2; uv[1] = 2;
			glVertexAttrib2fv(uvcoord, uv);
			glColor3f  (1.0,0.5,0.2);
			glVertex3d ( 50, 50,0);

			uv[0] = 0; uv[1] = 2;
			glVertexAttrib2fv(uvcoord, uv);
			glColor3f  (1.0,0.5,0.2);
			glVertex3d (-50, 50,0);
		glEnd ();
	}
	glPopMatrix ();

	/* use the custom shader program pointed to by the given handle,	  */
	/* until we call this function again to change it					  */	  
	glUseProgram (shaderProg[0]);

	// save a copy of the current model transformation state
	glPushMatrix ();
	{
		/* notes: matrix operations are post-multiplied onto the GL matrix stack. */

		/* this polygon is 200 units in front of the image */
		glTranslatef (0,-100,200);

		/* rotate about centroid of polygon, on x axis			    */
		/* note: must center polygon since origin isn't at centroid */
		glRotatef    (angle, 1,0,0);
		glTranslatef (0,-50,0);

		/* rainbow polygon */
		glColor3d (0,0,1);
		glBegin (GL_QUADS);
			glColor3f  (1,0,0);
			glVertex3d (-50,0,0);

			glColor3f  (0,1,0);
			glVertex3d ( 50,0,0);

			glColor3f  (0,0,1);
			glVertex3d ( 50,100,0);

			glColor3f  (1,1,0);
			glVertex3d (-50,100,0);
		glEnd ();
	}
	glPopMatrix ();

	/* use the custom shader program pointed to by the given handle,	  */
	/* until we call this function again to change it					  */	
	glUseProgram (shaderProg[1]);
	/* the image, drawns as pixels */
	glPushMatrix ();
	{
		glTranslatef (-image->width()/2.0,-image->height()/2.0,0);

		GLint uvcoord = glGetAttribLocation (shaderProg[1], "vertex_texture_coordinate");
		GLfloat uv[2];

		glColor3f (1.0,1.0,1.0);
		glBegin (GL_QUADS);
			uv[0] = 0; uv[1] = 0;
			glVertexAttrib2fv(uvcoord, uv);
			glVertex3d (0,0,0);

			uv[0] = 1; uv[1] = 0;
			glVertexAttrib2fv(uvcoord, uv);
			glVertex3d (image->width(),0,0);

			uv[0] = 1; uv[1] = 1;
			glVertexAttrib2fv(uvcoord, uv);
			glVertex3d (image->width(),image->height(),0);

			uv[0] = 0; uv[1] = 1;
			glVertexAttrib2fv(uvcoord, uv);
			glVertex3d (0,image->height(),0);
		glEnd ();
	}
	glPopMatrix ();

	/* use the default OpenGL pipeline functionality, until we call this  */
	/* function again to change it										  */
	glUseProgram (0);

	// draw a wirefqrame cube
	glPushMatrix ();
		glScalef (400,300,400);
		draw_cube ();
	glPopMatrix ();

	/* Complete any pending operations */
    glFlush();         

	/* Make the drawing buffer the frame buffer and vice versa */
    glutSwapBuffers(); 
}


/* -------------------------------------------------------------------------- */
/*
// This is the default background activity; this function is
// called whenever no events are detected.
*/
static void idle_CB (void)
{
	/* this is a crude way to regulate the pump 
	#if defined(WIN32)
		Sleep(100);
	#else
		sleep(1);
	#endif
	*/

	/* state has changed: tell GLUT to redraw. */
	glutPostRedisplay();
}


/* -------------------------------------------------------------------------- */
/* Function called is called on a key press */
static void key_CB(unsigned char key, int x, int y) 
{
    dbc(key);   /* Print debugging stuff on stderr. */
    dbi(x);     /* Print debugging stuff on stderr. */
    dbi(y);     /* Print debugging stuff on stderr. */

    switch (key)
	{		
		/* save the image raster back to file */
		case 's':
			write_jpeg_image (outfile,*image);
			break;

		/* reload the image */
		case 'l':
			delete (image);
			image = read_jpeg_image (infile);
			break;
		/* quit the program -- 'hard quit' */
		case 'q':
			exit(0);
			break;
		default: 
			break;
	}
	/* state has changed: tell GLUT to redraw. */
	glutPostRedisplay ();
}

